/*-------------------------------------------------------------------------
	Copyright (C) 2008-2009 American Megatrends Inc.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
	
	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-------------------------------------------------------------------------*/
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <asm/io.h>
#include <asm/irq.h>
#include "i2c-data.h"
#include "i2c-hardware.h"

#define DEBUG 
//#define DEBUG_VERBOSE 

/* Each transmission has so many ms to complete. 
 * This time is based on number of bytes transmitted
 * or received. This time is more than enough and has
 * been padded to consider start, stop, and repeated
 * start conditions as well as slow slave devices. */ 
#define INTERRUPT_TIMEOUT_PER_BYTE 	30  

/* I2C driver flag used for reseting mmux via gpio */
#define I2C_M_RESET_MUX		0x8000

/* Prototypes */
int vsc_master_xfer(struct i2c_adapter *i2c_adap, struct i2c_msg *msgs,	int num);
int vsc_slave_xfer(struct i2c_adapter *i2c_adap, char *buf, int	num);
int if_bus_bad_do_recovery(struct i2c_adapter *i2c_adap);

static u32 i2c_vsc_wait_event_timeout(int bus , int ms_timeout)
{
	wait_event_interruptible_timeout(vsc_data_ptr[bus].vsc_wait, 
						vsc_data_ptr[bus].op_status, msecs_to_jiffies(ms_timeout));
	return vsc_data_ptr[bus].op_status;
}

static int vsc_send_bytes(struct i2c_adapter *i2c_adap,
				struct i2c_msg msg, int messages_left)
{
	int i;
	u32 op_status;

	/* Set flag to Master Transfer Mode */
	vsc_data_ptr[i2c_adap->nr].state = STATE_MASTER_XMIT;

	/* Filling data onto buffer */
	for (i=0; i<msg.len; i++) {
		vsc_data_ptr[i2c_adap->nr].TX_data[i] = msg.buf[i];
	}
    
	/* Initialize operation status value */
	vsc_data_ptr[i2c_adap->nr].op_status = OP_STATUS_TIMEOUT;

	vsc_data_ptr[i2c_adap->nr].TX_len = msg.len;
	vsc_data_ptr[i2c_adap->nr].TX_index = 0;
    
	/* Set scan flag if just scanning */
	if (0 == msg.len)
		vsc_data_ptr[i2c_adap->nr].scan = 1;
	else
		vsc_data_ptr[i2c_adap->nr].scan = 0;		

	/* Make sure bus is still free before taking ownership */
	if (is_i2c_bus_busy(i2c_adap->nr)) {
#ifdef DEBUG
		printk(KERN_ERR "I2C%d:0x%02x: Bus busy (vsc_send_bytes) : bus_busy_cnt=%d dev_fail_cnt=%d\n", 
			i2c_adap->nr, msg.addr<<1, vsc_data_ptr[i2c_adap->nr].bus_busy_cnt,
			(int)vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[msg.addr]);
#endif
		return -1;
	}

	/* Clear stop detected status */
	i2c_vsc_write_reg((UINT32)0x40, (UINT32)i2c_vsc_io_base(i2c_adap->nr) + VSC_I2CX_BIT_STATUS);

	/* Enable I2C interface by sending START here*/
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS 
				| I2C_ENABLE_INTERFACE
				| I2C_ENABLE_INTERRUPT
				| I2C_SEND_START
				|I2C_ACKNOWLEDGEMENT),
				(UINT32)i2c_vsc_io_base(i2c_adap->nr) + VSC_I2CX_CNTRL);

	/* Waiting here for transmission to complete (rest of the logic handled inside interrupt handler) */
	op_status = i2c_vsc_wait_event_timeout(i2c_adap->nr, INTERRUPT_TIMEOUT_PER_BYTE * (msg.len + 1));

	/* Reinitialise I2C internal data and hardware */
	i2c_init_internal_data(i2c_adap->nr);
	i2c_init_hardware(i2c_adap->nr);

	/* If transmission was unsuccessful */
	if (OP_STATUS_SUCCESS != op_status) {


		/* If this was just an i2c device scan */
		if (vsc_data_ptr[i2c_adap->nr].scan) {

			/* If we found an i2c device */
			if (OP_STATUS_DEV_FOUND == op_status) {

				/* Return success */
				return 0;
			}
		}

		/* If NACK error */
		if (OP_STATUS_NACK == op_status) {
#ifdef DEBUG
			printk(KERN_ERR "I2C%d:0x%02x: Device NACK (vsc_send_bytes) : dev_fail_cnt=%d\n", 
				i2c_adap->nr, msg.addr<<1, (int)vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[msg.addr]);
#endif
			return -3; /* Return NACK failure */
		} else {
#ifdef DEBUG
			printk(KERN_ERR "I2C%d:0x%02x: Unsuccessful op_status=%d (vsc_send_bytes) : dev_fail_cnt=%d\n", 
				i2c_adap->nr, msg.addr<<1, (int)op_status, (int)vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[msg.addr]);
#endif
			return -2; /* Return operation status failure */
		}
	}

	/* Return success */
	return 0;
}

static int vsc_receive_bytes(struct i2c_adapter *i2c_adap,
			struct i2c_msg msg, int messages_left)
{
	int i;
	u32 op_status;

	vsc_data_ptr[i2c_adap->nr].MasterRX_index = 0;
	vsc_data_ptr[i2c_adap->nr].MasterRX_len = msg.len;

	vsc_data_ptr[i2c_adap->nr].TX_len = 0;
    
	/* Set flag to Master Receive Mode */
	vsc_data_ptr[i2c_adap->nr].state = STATE_MASTER_RECV;

	/* Initialize operation status value */
	vsc_data_ptr[i2c_adap->nr].op_status = 0;

	/* Make sure bus is still free before taking ownership (i.e. issuing start condition) */
	if (is_i2c_bus_busy(i2c_adap->nr)) {
#ifdef DEBUG
		printk(KERN_ERR "I2C%d:0x%02x: Bus busy (vsc_receive_bytes) : bus_busy_cnt=%d dev_fail_cnt=%d\n", 
			i2c_adap->nr, msg.addr<<1, vsc_data_ptr[i2c_adap->nr].bus_busy_cnt,
			(int)vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[msg.addr]);
#endif
		return -1;
	}

	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS 
				| I2C_ENABLE_INTERFACE
				| I2C_ENABLE_INTERRUPT
				| I2C_SEND_START
				| I2C_ACKNOWLEDGEMENT),
				(UINT32)i2c_vsc_io_base(i2c_adap->nr) + VSC_I2CX_CNTRL);
    
	/* Waiting here for transmission to complete (rest of the logic handled inside interrupt handler) */
	op_status = i2c_vsc_wait_event_timeout(i2c_adap->nr, INTERRUPT_TIMEOUT_PER_BYTE * (msg.len + 1));

	/* Reinitialise I2C internal data and hardware */
	i2c_init_internal_data(i2c_adap->nr);
	i2c_init_hardware(i2c_adap->nr);

	/* If transmission was unsuccessful */
	if (OP_STATUS_SUCCESS != op_status) {



		/* If NACK error */
		if (OP_STATUS_NACK == op_status) {
#ifdef DEBUG
			printk(KERN_ERR "I2C%d:0x%02x: Device NACK (vsc_receive_bytes) : dev_fail_cnt=%d\n", 
				i2c_adap->nr, msg.addr<<1, (int)vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[msg.addr]);
#endif
			return -3; /* Return NACK failure */
		} else {
#ifdef DEBUG
			printk(KERN_ERR "I2C%d:0x%02x: Unsuccessful op_status=%d (vsc_receive_bytes) : dev_fail_cnt=%d\n", 
				i2c_adap->nr, msg.addr<<1, (int)op_status, (int)vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[msg.addr]);
#endif
			return -2; /* Return operation status failure */
		}
	}
    
	/* Getting back response data from the buffer */
	for (i=0; i<msg.len; i++)
		msg.buf[i] = vsc_data_ptr[i2c_adap->nr].MasterRX_data[i];

	/* Return success */
	return 0;
}

static int vsc_send_receive_bytes(struct i2c_adapter *i2c_adap,
					struct i2c_msg write_msg, struct i2c_msg read_msg)
{
	int i;
	u32 op_status;

	/* Filling data onto buffer */
	for (i=0; i<write_msg.len; i++) {
		vsc_data_ptr[i2c_adap->nr].TX_data[i] = write_msg.buf[i];
	}
    
	vsc_data_ptr[i2c_adap->nr].TX_len = write_msg.len;
	vsc_data_ptr[i2c_adap->nr].TX_index = 0;
   
	vsc_data_ptr[i2c_adap->nr].MasterRX_index = 0;
	vsc_data_ptr[i2c_adap->nr].MasterRX_len = read_msg.len;

	/* Set scan flag if just scanning */
	if (0 == write_msg.len)
		vsc_data_ptr[i2c_adap->nr].scan = 1;
	else
		vsc_data_ptr[i2c_adap->nr].scan = 0;		

	/* Initialize operation status value */
	vsc_data_ptr[i2c_adap->nr].op_status = OP_STATUS_TIMEOUT;

	/* Set flag to Master Transfer Mode */
	vsc_data_ptr[i2c_adap->nr].state = STATE_MASTER_XMIT_RECV;

	/* Make sure bus is still free before taking ownership */
	if (0 != is_i2c_bus_busy(i2c_adap->nr)) {

#ifdef DEBUG
		printk(KERN_ERR "I2C%d:0x%02x: Bus busy (vsc_send_receive_bytes) : bus_busy_cnt=%d dev_fail_cnt=%d \n", 
			i2c_adap->nr, write_msg.addr<<1, vsc_data_ptr[i2c_adap->nr].bus_busy_cnt, vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[write_msg.addr]);
#endif
		return -1;
	}

	/* Clear stop detected status */
	i2c_vsc_write_reg((UINT32)0x40, (UINT32)i2c_vsc_io_base(i2c_adap->nr) + VSC_I2CX_BIT_STATUS);

	/* Enable I2C interface by sending START here*/
	i2c_vsc_write_reg((UINT32)(I2C_CLEAR_ALL_STATUS 
				| I2C_ENABLE_INTERFACE
				| I2C_ENABLE_INTERRUPT
				| I2C_SEND_START
				|I2C_ACKNOWLEDGEMENT),
				(UINT32)i2c_vsc_io_base(i2c_adap->nr) + VSC_I2CX_CNTRL);

	/* Waiting here for transmission to complete (rest of the logic handled inside interrupt handler) */
	op_status = i2c_vsc_wait_event_timeout(i2c_adap->nr, INTERRUPT_TIMEOUT_PER_BYTE * ((write_msg.len + 2) + (read_msg.len + 1)));

	/* Reinitialise I2C internal data and hardware */
	i2c_init_internal_data(i2c_adap->nr);
	i2c_init_hardware(i2c_adap->nr);

	/* If transmission was unsuccessful */
	if (OP_STATUS_SUCCESS != op_status) {

		/* If this was just an i2c device scan */
		if (vsc_data_ptr[i2c_adap->nr].scan) {

			/* If we found an i2c device */
			if (OP_STATUS_DEV_FOUND == op_status) {

				/* Return success */
				return 0;
			}
		}

		/* If NACK error */
		if (OP_STATUS_NACK == op_status) {
#ifdef DEBUG
			printk(KERN_ERR "I2C%d:0x%02x: Device NACK (vsc_send_receive_bytes) : dev_fail_cnt=%d\n", 
				i2c_adap->nr, write_msg.addr<<1, (int)vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[write_msg.addr]);
#endif
			return -3; /* Return NACK failure */
		} else {
#ifdef DEBUG
			printk(KERN_ERR "I2C%d:0x%02x: Unsuccessful op_status=%d (vsc_send_receive_bytes) : dev_fail_cnt=%d\n", 
				i2c_adap->nr, write_msg.addr<<1, (int)op_status, (int)vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[write_msg.addr]);
#endif
			return -2; /* Return operation status failure */
		}
	}

	/* Copy read I2C data to buffer */
	for (i=0; i<read_msg.len; i++)
		read_msg.buf[i] = vsc_data_ptr[i2c_adap->nr].MasterRX_data[i];

	/* Return success */
	return 0;
}


static int send_slave_addr(struct i2c_adapter *i2c_adap, int addr, int flags)
{
	/* 
	 * Writting slave address onto data register 
	 * Slave address must be left shifted by 1 bit so as to append the R/W bit
	 */
	vsc_data_ptr[i2c_adap->nr].slave_addr = addr;

	if (flags & I2C_M_RD) {
		/* Append R/W bit as 1(one) */
		i2c_vsc_write_reg((addr << 1) | READ_OPERATION_BIT, 
			(UINT32)i2c_vsc_io_base(i2c_adap->nr) + VSC_I2CX_DATA);
	} else {
		/* Append R/W bit as 0(zero) */
		i2c_vsc_write_reg((addr << 1) | WRITE_OPERATION_BIT, 
			(UINT32)i2c_vsc_io_base(i2c_adap->nr) + VSC_I2CX_DATA);
	}

	return 0;
}


int vsc_master_xfer(struct i2c_adapter *i2c_adap,
			struct i2c_msg *msgs, int num)
{
	int i = 0;
	int retval = 0;

#ifdef DEBUG_VERBOSE
	printk(KERN_ERR "vsc_master_xfer: bus=%d ", i2c_adap->nr);
	for (i=0; i<num-1; i++)
		printk("msgs[%d].addr=0x%02x, ", i, msgs[i].addr<<1);
	printk("msgs[%d].addr=0x%02x\n", num-1, msgs[num-1].addr<<1);
#endif

	/* If too many consecutive bus busy failures have occured on this bus, it might be hung */
	if (vsc_data_ptr[i2c_adap->nr].bus_busy_cnt >= MAX_BUS_BUSY_CNT) {

		/* Is the I2C bus hung? */
		if (vsc_is_i2c_bus_hung(i2c_adap)) {

			/* Perform I2C bus recovery */
			if (0 != vsc_i2c_bus_recovery(i2c_adap)) {

				/* Reinitialise I2C internal data and hardware */
				i2c_init_internal_data(i2c_adap->nr);
				i2c_init_hardware(i2c_adap->nr);
				return -1;
			}
		}
	}

	/* Check if bus busy */ 
	if (i2c_wait_for_bus_free(i2c_adap->nr)) {
#ifdef DEBUG
		printk(KERN_ERR "I2C%d: Bus busy (vsc_master_xfer) : bus_busy_cnt=%d\n", i2c_adap->nr, vsc_data_ptr[i2c_adap->nr].bus_busy_cnt);
#endif
		/* Reinitialise I2C internal data and hardware */
		i2c_init_internal_data(i2c_adap->nr);
		i2c_init_hardware(i2c_adap->nr);
		return -1;
	}
    
	for (i=0; ((retval>=0) && (i<num)); i++) {

		/* Disallow access to our own slave address */
		if ((u8)msgs[i].addr == (u8)(slave_addr[i2c_adap->nr]))
			return -1;

		/* Setup slave address */
		if (msgs[i].flags & I2C_M_RD) {
			send_slave_addr(i2c_adap, msgs[i].addr, READ_OPERATION_BIT);
		} else {
			send_slave_addr(i2c_adap, msgs[i].addr, WRITE_OPERATION_BIT);
		} 

		/* If this msg is a write operation */
		if ((msgs[i].flags & I2C_M_RD) == 0) {
			/* If another operation follows this write operation */
			if ((i + 1) < num) {
				/* If the next operation is a read operation to same slave address */
				if ((msgs[i+1].flags & I2C_M_RD) && (msgs[i+1].addr == msgs[i].addr)) {
					/* Perform write operation followed by repeated start followed by read operation */
					retval = vsc_send_receive_bytes(i2c_adap, msgs[i], msgs[i+1]);
					/* Increment i one additional time since we just performed two operations */
					i++;
				} else {
					/* Otherwise perform a single master write operation */
					retval = vsc_send_bytes(i2c_adap, msgs[i], num);
				}
			} else {
				/* Otherwise perform a single master write operation */
				retval = vsc_send_bytes(i2c_adap, msgs[i], num);
			}
		/* If this is a read operation */
		} else {
			/* Perform a single master read operation */
			retval = vsc_receive_bytes(i2c_adap, msgs[i], num);
		} 

		/* If any operation failed, return error */
		if (0 != retval) {

			/* Keep track of consecutive failures per device */
			vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[msgs[i].addr]++; 

			if (vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[msgs[i].addr] >= MAX_FAILURES_PER_DEVICE) {
				/* Is the I2C bus hung? */
				if (vsc_is_i2c_bus_hung(i2c_adap)) {

					/* Perform I2C bus recovery */
					vsc_i2c_bus_recovery(i2c_adap);
				}
				/* Reset device error count after bus recovery */
				vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[msgs[i].addr] = 0; 
			}

			/* Reinitialise I2C internal data and hardware */
			i2c_init_internal_data(i2c_adap->nr);
			i2c_init_hardware(i2c_adap->nr);

			/* Reset all muxes if so specified */
			if (msgs[i].flags & I2C_M_RESET_MUX)
				reset_mux_via_gpio(i2c_adap->nr);		

			return retval;
		}
		else {
			/* Reset device error count since we successfully accessed device */
			vsc_data_ptr[i2c_adap->nr].dev_fail_cnt[msgs[i].addr] = 0; 
		}

		/* Reset all muxes if so specified */
		if (msgs[i].flags & I2C_M_RESET_MUX)
			reset_mux_via_gpio(i2c_adap->nr);		
	}

	/* Successful transmission occured so reset bus busy counter */
	vsc_data_ptr[i2c_adap->nr].bus_busy_cnt = 0; 

	return num;
}

int vsc_slave_xfer(struct i2c_adapter *i2c_adap,
			char *buf, int num)
{
	/* Currently not implemented */
	return 0;
}
