
/*****************************************************
 *   NAND Suppport for VSC452 
 *
 * (C) Copyright 2008 American Megatrends Inc
 *****************************************************/

#include <common.h>

#if (CONFIG_COMMANDS & CFG_CMD_NAND)

#include <nand.h>

#define NAND_RD_DATA(base)		*((volatile unsigned long *)((base)+0x00))
#define NAND_WR_DATA(base)		*((volatile unsigned long *)((base)+0x04))
#define NAND_ADDR(base)			*((volatile unsigned long *)((base)+0x08))
#define NAND_COMMAND(base)		*((volatile unsigned long *)((base)+0x0C))
#define NAND_CONTROL(base)		*((volatile unsigned long *)((base)+0x10))
#define NAND_STATUS(base)		*((volatile unsigned long *)((base)+0x14))

#define STATUS_READY			0x02


/* VSC452 does not have support to control signals. It does 
   automatically the command and address register are written */
static u8 hwctl = 0;

#define COMMAND_PHASE 0x01
#define ADDR_PHASE	  0x02

static void 
nand_hwcontrol(struct mtd_info *mtdinfo, int cmd)
{
	switch (cmd) 
	{
		case NAND_CTL_SETCLE:
			hwctl |= COMMAND_PHASE;
			break;

		case NAND_CTL_CLRCLE:
			hwctl &= ~COMMAND_PHASE;
			break;

		case NAND_CTL_SETALE:
			hwctl |= ADDR_PHASE;
			break;

		case NAND_CTL_CLRALE:
			hwctl &= ~ADDR_PHASE;
			break;

	case NAND_CTL_SETNCE:
	case NAND_CTL_CLRNCE:
	default:
		break;
	}
	return;
}

static void 
nand_write_byte(struct mtd_info *mtdinfo, u_char byte)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_W; 

	if (hwctl & COMMAND_PHASE)
		NAND_COMMAND(base) = byte;
	else 
	{
		if (hwctl & ADDR_PHASE)
			NAND_ADDR(base) = byte;
		else
			NAND_WR_DATA(base) = byte;
	}
}

static u_char 
nand_read_byte(struct mtd_info *mtdinfo)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_R;

	return NAND_RD_DATA(base);
}

static int
nand_dev_ready(struct mtd_info *mtdinfo)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_R;

	if  (!(NAND_STATUS(base) & STATUS_READY))
		return 0;
	return 1;
}


static void 
nand_read_buf(struct mtd_info *mtdinfo, uint8_t *buf, int len)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_R;

	for (;len > 0; len --)
		*buf++ = NAND_RD_DATA(base);
	return;
}

static void 
nand_write_buf(struct mtd_info *mtdinfo, const uint8_t *buf, int len)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_W;

	for (; len > 0; len --)
			NAND_WR_DATA(base) =  *buf++;
}

static int 
nand_verify_buf(struct mtd_info *mtdinfo, const uint8_t *buf, int len)
{
	struct nand_chip *this = mtdinfo->priv;
	ulong base = (ulong) this->IO_ADDR_R;

	for (; len > 0; len --)
		if (*buf++ !=NAND_RD_DATA(base))
			return -1;

	return 0;
}


void 
board_nand_init(struct nand_chip *nand)
{
	ulong base = (ulong) nand->IO_ADDR_W;

	nand->hwcontrol  = nand_hwcontrol;
	nand->read_byte  = nand_read_byte;
	nand->write_byte = nand_write_byte;
	nand->dev_ready  = nand_dev_ready;
	nand->eccmode    = NAND_ECC_SOFT;
	nand->write_buf  = nand_write_buf;
	nand->read_buf   = nand_read_buf;
	nand->verify_buf = nand_verify_buf;

	/* Enable Chip Select CS0 */
	NAND_CONTROL(base) = 0x10;
	return;
}

#endif
